/*
   localf.cc -- local_finish
   
   (c) 1993, 1994, 1995 Han-Wen Nienhuys <hanwen@stack.urc.tue.nl>

   this is part of Rayce, a raytracer based on "Ray" by George Kyriazis	
   <kyriazis@turing.cs.rpi.edu> Rayce is released under the terms of the
   GNU General Public License, it has no warranty what so ever. For more
   details: see the file "COPYING"
*/

#include "texture.hh"
#include "proto.hh"
/****************************************************************
   Calculate various microfacet distributions

 * Some stuff below is based on the sources by Roy Hall:
 *
 * "Illumination and Color in Computer Generated Imagery" Source for
 * code given in the Appendices (c)Roy Hall - 1988,1989
 *
 *
 * I am making this code available to the user community.  You are
 * free to use it as you wish.  In return I request two things.
 * First, that this README file always stays with any copy of the
 * source, more specifically, that proper credit is given to the
 * original author. Second, that any corrections, improvements,
 * etc. are passed back to the author for inclusion in subsequent
 * distribution and editions of the book.

   I copied a small part of the D.c file, and adapted it to my shading
   code. Including the whole README seemed a bit of overkill

 ****************************************************************/

static inline flt
hdotn(vector ldir, vector raydir, vector normal)
{
    vector viewdir(raydir/(-raydir.length()) );
    vector halfway = ldir + viewdir;
    halfway /= halfway.length();
    flt retval= halfway * normal;
    assert (abs(retval) <= 1.0);
    return retval;
}

flt
phong_distribution(flt invroughness, vector ldir, vector raydir, vector normal)
{
    /* Calculate the reflected vector. */
    vector refl = 2 * (ldir * normal) * normal;
    refl -= ldir;
    vector rdotv = refl * (-raydir/raydir.length());

    /* Initialize factor. */
    if (rdotv > EPSILON)
	return pow(rdotv, invroughness);
    else return 0;
}

flt
cook_distribution(flt invroughness, vector ldir, vector raydir, vector normal)
{
    flt         c = sqr(hdotn(ldir,raydir,normal));
    flt	    r = sqr(invroughness);
    flt 	exponent = -((1 / c) - 1) / r;
    c = c * c;
    return exp(exponent) / (c * 4 * M_PI * r);
}

flt
reitz_distribution(flt invroughness, vector ldir, vector raydir, vector normal)
{
    flt c = sqr(hdotn(ldir,raydir,normal));
    flt r = sqr(1/invroughness);

    /* there is an error in the book "Image Synthesis" by Thalmann & Magenat-Thalman */
    return sqr(r / (c * (r - 1) + 1));
}

flt
gauss_distribution(flt invroughness, vector ldir, vector raydir, vector normal)
{
    flt alpha = acos(hdotn(ldir,raydir,normal));
    flt        exponent = -sqr((alpha * invroughness));
    return exp(exponent);
}

flt
blinn_distribution(flt invroughness, vector ldir, vector raydir, vector normal)
{
    flt hdn = hdotn(ldir,raydir,normal);

    return (hdn > EPSILON)? pow(hdn, invroughness) :0.0;
}

/****************************************************************
  find roughness from "halfintensity" angle
 ****************************************************************/

flt phong_setroughness(flt beta)
{
    if (beta < EPSILON || beta >= M_PI_4)
        return 0;


    else {
        return 1 / (-(log(2.0) / log(cos(2.0 * beta))));
    }
}

flt blinn_setroughness(flt beta)
{
    if (beta < EPSILON || beta > M_PI / 2)
        return 0;

    else {
        return 1 / (-(log(2.0) / log(cos(beta))));
    }
}

flt gauss_setroughness(flt beta)
{
    if (beta < EPSILON || beta > M_PI / 2)
        return 0;
    else {
	return 1 / (sqrt(log(2.0)) / beta);
    }
}

flt reitz_setroughness(flt beta)
{
    if (beta < EPSILON || beta > M_PI / 2)
        return 0;
    else    {
	flt         cos_beta = cos(beta);
	return sqrt((sqr(cos_beta) - 1.0) /
		    (sqr(cos_beta) - sqrt(2.0)));
    }
}

flt cook_setroughness(flt beta)
{
    if (beta < EPSILON || beta > M_PI / 2)
        return 0;
    else {
        flt tan_beta;

        tan_beta = tan(beta);
        return sqrt(-sqr(tan_beta)  /
			   log(pow(cos(beta), 4.0) / 2.0));
    }
}




/****************************************************************
  GEOMETRIC
 ****************************************************************/

flt nogeo(vector , vector , vector ){
    return 1.0;
}

flt torrance(vector ldir, vector raydir, vector normal){
//    flt ldotn = ldir * normal;

    vector viewdir = raydir;
    viewdir /= -viewdir.length();

    vector H = ldir + viewdir;
    H /= H.length();

    flt hn = H * normal;
    flt hl = H * ldir;
    if (hl < EPSILON)
        return 1.0;
    flt gm = 2 * hn * (normal * ldir) / hl;
    flt gs = 2 * hn * (viewdir * normal) / hl;
    flt g = MIN(gm, gs);
    g = MIN(1.0, g);
    return g;
}
void
local_finish::set_rough(double a)
{
    roughness =  set_rough_func(a);
}

void
local_finish::set_distr(int d)
{
    dtype = distributions(d);
    switch (dtype) {
    case DPHONG:
	set_rough_func = phong_setroughness;
	distfunc = phong_distribution;
	break;
    case DGAUSS:
	set_rough_func = gauss_setroughness;
	distfunc = gauss_distribution;
	break;
     case DBLINN:
	set_rough_func = blinn_setroughness;
	distfunc = blinn_distribution;
	break;
    case DCOOK:
	set_rough_func = cook_setroughness;
	distfunc = cook_distribution;
	break;
    case DREITZ:
	set_rough_func = reitz_setroughness;
	distfunc = reitz_distribution;
	break;
    default:
	errormsg("don't know distribution %d", d);
    }
}


void
local_finish::set_geo(int d)
{
    gtype = geometric(d);
    switch (d) {
    case GEO_NONE:
	geofunc = nogeo;
	break;
    case GEO_TORRANCE:
	geofunc = torrance;
	break;
    default:
	errormsg("don't know geometric %d", d);
    }
}

void
local_finish::print() const
{
#ifdef DEBUG
    
    dprintf("local finish {micro %d, geo %d\n", dtype, gtype);
    dprintf("ambient");
    ambient.print();
    dprintf("diffuse");
    diffuse.print();
    dprintf("brilliance %f\n", brilliance);

    dprintf("specular");
    specular.print();
    dprintf("roughness %f\n", roughness);

    dprintf("reflection (ang,col) %f ", refl_diffuse);
    refl_color.print();

    dprintf("refraction (ior,angle,color) %f %f, ", index, refr_diffuse);
    refr_color.print();
    dprintf("}\n");
#endif
}

local_finish::local_finish() :
   refl_color(0), refr_color(0), ambient(0), diffuse(0), specular(0)
{
    set_distr(DBLINN);
    set_geo(GEO_NONE);
    index = 1.0;
    refl_diffuse =refr_diffuse = 0.0;
    brilliance =1.0;
    roughness =0.0;
}
