/*
   Raymath.cc -- simple math routines.
 *
 * based on work by George Kyriazis, 1988

 (c) 1993, 1994, 1995 Han-Wen Nienhuys <hanwen@stack.urc.tue.nl>

 this is part of Rayce, a raytracer based on "Ray" by George Kyriazis
 <kyriazis@turing.cs.rpi.edu> Rayce is released under the terms of the
 GNU General Public License, it has no warranty what so ever. For more
 details: see the file "COPYING"
 */

/*
 * math and vector operations
 */

#include	"ray.hh"
#include	"proto.hh"
#include	"extern.hh"
#include	"random.hh"
#include	"matrix.hh"

/*
 * floating point random number generator, from K&R chap 7.8.7 [deleted]
 * now with table lookup.
 */

PRIVATE int randidx;
// between 0 and 1
flt
  frand()
{
    if (--randidx < 0)
        randidx = rand() % RANDOMTABSIZE;

    return randtable[randidx];
}


/*
 * Random number between -1 and 1
 */
  flt
        rand1()
{
    if (--randidx < 0)
        randidx = rand() % RANDOMTABSIZE;

    return (randtable[randidx] - 0.5) * 2;
}
/*
 * approximate a gaussian random number generator between -1 and 1
 */
PUBLIC  flt
        Gauss_rand()
{
    flt     t;
    /* get a random number between -1 and 1 */
    t = rand1();

    /* and then square it.  Don't lose the sign! */
    return (t * t * SGN(t));
}
/*
   Performs a safe arctangent calculation for the two provided numbers.
   If the denominator is 0, atan2 is not called (it results in a
   floating exception on some machines).  Instead, + or - PI/2 is
   returned.
   */

PUBLIC  flt
        safe_arctangent(flt y, flt x)
{
    if (x != 0.0)
        return atan2(y, x);
    else if (y > 0.0)
        return M_PI_2;
    else
        return -M_PI_2;
}
/************************** VECTOR ********************************/

axes
nc_vector::dominant() const
{
    axes d =X;
    flt m = abs(x);
    if ( abs(y) > m) {
	d = Y;
	m = abs(y);
    }
    if ( abs(z) > m) {
	d = Z;
    }
    return d;
}

void
nc_vector::print(const char *s) const
{
#ifdef DEBUG
    dprintf("%s: <%f, %f, %f>\n", s, x, y, z);
#endif          /* DEBUG */
}

void
vector::rotate(const matrix & M)
{
    *this = M * (*this);
}

/***************************************************************
  MATRIX
 ***************************************************************/

/* ?rotate() generates matrix for lefthanded rotation about ? axis. */
PRIVATE void
xrotate(matrix  & m, const flt angle)
{
    m.unit();
    m[1][1] = m[2][2] = cos(angle);
    m[1][2] = -sin(angle);
    m[2][1] = -m[1][2];
}
PRIVATE void
yrotate(matrix  & m, const flt angle)
{
    m.unit();
    m[0][0] = m[2][2] = cos(angle);
    m[0][2] = sin(angle);
    m[2][0] = -m[0][2];
}

PRIVATE void
zrotate(matrix  & m, const flt angle)
{
    m.unit();
    m[0][0] = m[1][1] = cos(angle);
    m[1][0] = sin(angle);
    m[0][1] = -m[1][0];
}

void
matrix::rotate(const vector r)
{
    matrix  A,B;
    xrotate(A, r.x);
    yrotate(B, r.y);
    A = B * A;
    zrotate(B, r.z);
    *this = B * A;
}

void
matrix::make_rotation(const vector v)
{
    vector d;

    d.x = degtorad(v.x);
    d.y = degtorad(v.y);
    d.z = degtorad(v.z);

    rotate(d);
}

void
matrix::print() const
{
#ifdef DEBUG
    int     i,
            j;
    dprintf("\n");
    for (i = 0; i < 4; i++)
    {
        for (j = 0; j < 4; j++)
            dprintf("%10lf ", (*this)[i][j]);
        dprintf("\n");
    }
#endif          /* DEBUG */
}

matrix  matrix::transpose() const
{
    matrix  tr;
    int     i,
            j;
    for (i = 0; i < 4; i++)
        for (j = 0; j < 4; j++)
            tr[i][j] = (*this)[j][i];
    return tr;
}

void
matrix::inv_translate(const vector t)
{

    translate(-t);
}

/* translate_matrix */
void
matrix::translate(const vector t)
{
    unit();
    (*this)[0][3] = t.x;
    (*this)[1][3] = t.y;
    (*this)[2][3] = t.z;
}

void
matrix::inv_rotate(const vector r)
{
    rotate(r);
    transpose();
}
/* generate a matrix for scaling with the vector s */

void
matrix::scale(const vector s)
{
    unit();
    (*this)[0][0] = s.x;
    (*this)[1][1] = s.y;
    (*this)[2][2] = s.z;
}

void
matrix::inv_scale(const vector s)
{
    unit();
    (*this)[0][0] = 1 / s.x;
    (*this)[1][1] = 1 / s.y;
    (*this)[2][2] = 1 / s.z;
}

void
matrix::diag(const flt d)
{
    int     i,
            j;
    for (i = 0; i < 4; i++)
        for (j = 0; j < 4; j++)
	    (*this)[i][j] =  (i == j)  ? d: 0;

}


/*
 * multiply matrix a with matrix
 */
matrix operator *(const matrix & d, const matrix & b)
{
    matrix  c;
    int     i,
            j,
            k;


    for (i = 0; i < 4; i++)
        for (j = 0; j < 4; j++)
        {
            c[i][j] = 0;
            for (k = 0; k < 4; k++)
                c[i][j] += d[i][k] * b[k][j];
        }
    return c;
}
/*
 * matrix vector product: multiply  a 3x3 (or 4x4) matrix with
 * columnvector x
 */
vector operator *(const matrix & A, const vector & x)
{
    vector  v;
    v.x = A[0][0] * x.x + A[0][1] * x.y + A[0][2] * x.z + A[0][3];
    v.y = A[1][0] * x.x + A[1][1] * x.y + A[1][2] * x.z + A[1][3];
    v.z = A[2][0] * x.x + A[2][1] * x.y + A[2][2] * x.z + A[2][3];

    return v;
}
/* multiply with transpose of A */
vector
transform_normal(const matrix & A, const vector & n)
{
    vector  v;
    v.x = n.x * A[0][0] + n.y * A[1][0] + n.z * A[2][0];
    v.y = n.x * A[0][1] + n.y * A[1][1] + n.z * A[2][1];
    v.z = n.x * A[0][2] + n.y * A[1][2] + n.z * A[2][2];

    return v;
}
/*
 * matrix-vector multiply, but don't use translation.
 */
vector operator *(const matrix & A, const direction & x)
{
    vector  v;
    v.x = A[0][0] * x.x + A[0][1] * x.y + A[0][2] * x.z;
    v.y = A[1][0] * x.x + A[1][1] * x.y + A[1][2] * x.z;
    v.z = A[2][0] * x.x + A[2][1] * x.y + A[2][2] * x.z;

    return v;
}
/*
 * matrix-vector multiply, but don't use translation.
 */
vector  transform_direction(const matrix & A, const vector & x)
{
    vector  v;
    v.x = A[0][0] * x.x + A[0][1] * x.y + A[0][2] * x.z;
    v.y = A[1][0] * x.x + A[1][1] * x.y + A[1][2] * x.z;
    v.z = A[2][0] * x.x + A[2][1] * x.y + A[2][2] * x.z;

    return v;
}

