/*
  shapes.hh --type definitions for primitives and nonprimitives.

 (c) 1993, 1996 Han-Wen Nienhuys <hanwen@stack.urc.tue.nl>


 */


#ifndef SHAPES_H
#define SHAPES_H
#include "stats.hh"
#include "object.hh"
#include "shapeheads.hh"

/*@ManMemo: quadric equation coefficients */
struct quadric_data : shape_data {
    MAKESHAPE(quadric);

    vector  normal(vector) const;
    bool    inside(vector) const;
    void rotate(matrix const & rotmat);
    void translate(const vector t);
    void scale(const vector s);
    quadric_data(vector a1, vector a2, vector a3, flt a4);

private:
    vector  qt;
    vector   bqt;
    vector  lt;                 /*@ManMemo: quadratic, biquadratic, linear terms */
    flt     a;
    int do_intersection(ray * r, flt *, int flags) const ;
};


struct sphere_data : shape_data {
    MAKESHAPE(sphere);
    void bound(matrix *,bounding_box *m);
    sphere_data(vector ,flt);    
    vector  normal(vector) const;
    bool    inside(vector) const;
    void rotate(matrix const & rotmat);
    void translate(const vector t);
    void scale(const vector s);
    virtual int do_intersection(ray * r, flt *, int flags) const ;
    vector  center;
    flt     radius;
protected:
    flt rsq;                    /*@ManMemo: radius squared */
};

/// (capped) finite cone
struct cone_data : shape_data {
    MAKESHAPE(cone);
    void bound(matrix *,bounding_box *m);

    vector  p1;
    vector p2;    
    flt     r1;
    flt r2;
    bool    locap,
    hicap;              /*@ManMemo: are bottom/top to be capped ? */

    cone_data(vector,vector,flt,flt);
    vector  normal(vector) const;
    bool    inside(vector) const;
    void rotate(matrix const & rotmat);
    void translate(const vector t);
    void scale(const vector s);
private:
    vector axis;
    flt r1sq,    r2sq,

    slope,
    axlen;
    
    int do_intersection(ray * r, flt *, int flags) const ;
};

///
struct plane_data : shape_data {
    MAKESHAPE(plane);

    vector  normal(vector) const;
    bool    inside(vector) const;
    void rotate(matrix const & rotmat);
    void translate(const vector t);
    void scale(const vector s);
    plane_data(vector, flt);
private:
    vector  n;                  /*@ManMemo: the normal */
    flt     mov;                /*@ManMemo: translation parallel to normal */
    int do_intersection(ray * r, flt *, int flags) const ;
};

struct disc_data : shape_data {
    MAKESHAPE(disc);
    void bound(matrix *,bounding_box *m);

    vector  n;                  /*@ManMemo: the normal */   
    vector  center;
    flt     majr,               /* radii */
    minr;
    vector  normal(vector) const;
    void rotate(matrix const & rotmat);
    void translate(const vector t);
    void scale(const vector s);
    disc_data(vector,vector,flt,flt);
bool all_intersections(ibuf * q, ray * r, int flags, bool & b) const;
private:
    flt     mov,    Rsq,                /* radii squared */
    rsq;                /* translation parallel to normal */
};

/*@ManMemo: a rectangular box */
struct box_data : shape_data {
    MAKESHAPE(box);
    vector  p1,
    p2;                 /*@ManMemo: point with minimum/maximum coordinates */
    void bound(matrix *,bounding_box *m);
    box_data(vector,vector);
    vector  normal(vector) const;
    bool    inside(vector) const;
    void translate(const vector t);
    void scale(const vector s);
private:
    int do_intersection(ray * r, flt *, int flags) const ;
};
struct torus_data : shape_data {
    MAKESHAPE(torus);

    flt brad; //bounding
	    flt Rad;
	flt rad;                            /*@ManMemo: major/minor radius */
    void bound(matrix *,bounding_box *m);
    torus_data(flt,flt);
    vector  normal(vector) const;
    int do_intersection(ray * r, flt *, int flags) const ;
    bool    inside(vector) const;

};
struct superq_data : shape_data {
    MAKESHAPE(superq);
    
    void bound(matrix *,bounding_box *m);
    superq_data(vector);
    vector  powvect;            /*@ManMemo: power of norm ||.|| */
    vector  normal(vector) const;
    int do_intersection(ray * r, flt *, int flags) const ;
    bool    inside(vector) const;
};
#endif

#if 0
struct extrusion_data : shape_data {
    MAKESHAPE(extrusion);
    object *shape;              /*@ManMemo: the shape to be extruded */
    flt     height;             /*@ManMemo: height of the extrusion */
    bool    locap,
    hicap;              /*@ManMemo: are bottom/top to be capped ? */

    ~extrusion_data();
    extrusion_data(extrusion_data &);
bool all_intersections(ibuf * q, ray * r, int flags, bool & b) const;
    vector  normal(vector) const;
    bool    inside(vector) const;
};

struct conic_data : shape_data {
    MAKESHAPE(conic);

    object *shape;              /*@ManMemo: the shape to be extruded */
    flt     height;             /*@ManMemo: height of the conic extrusion */
    vector  apex;               /*@ManMemo: apex of cone */
    bool    locap,
    hicap;              /*@ManMemo: are bottom/top to be capped ? */

bool all_intersections(ibuf * q, ray * r, int flags, bool & b) const;
    ~conic_data();
    conic_data(conic_data  &);
    vector  normal(vector) const;
    bool    inside(vector) const;

};
#endif

