/*
   sphere.c --  sphere routines.
   (c) 1993, 1994, 1995 Han-Wen Nienhuys <hanwen@stack.urc.tue.nl>

   this is part of Rayce, a raytracer based on "Ray" by George Kyriazis
   <kyriazis@turing.cs.rpi.edu> Rayce is released under the terms of the
   GNU General Public License, it has no warranty what so ever. For more
   details: see the file "COPYING"

   */


#include "matrix.hh"
#include "object.hh"
#include "shapes.hh"
#include "proto.hh"
#include "extern.hh"
#include "ibuf.hh"


shape_stats *rawsphere = shape_stats::regist("raw sphere", -1);

/* Intersect ray with sphere */
int
        intersect_sphere(vector center, flt rsq, ray * r, flt * hits)
{
    rawsphere->test++;

    /* solve: x = (pos+t*dir) and |x-center|^2  = radius^2 */
    vector      center_direction = r->pos - center; /* vector from center to r.pos */
    flt    dir_projection = r->dir * center_direction;
    flt    cd_lensq = center_direction * center_direction;     /* square of length of center_direction */

    if (cd_lensq >= rsq &&      /* we're not inside the sphere */
        dir_projection > 0)
    {                           /* and we're not looking at the sphere */
        return 0;
    }
    flt    linear = 2 * dir_projection;
    flt  constant = cd_lensq - rsq;

    /* with the advent of extrusions, |r.dir| isn't always 1 */
    flt    quadratic = r->dir * r->dir;
    flt  smallzero = constant / linear;


    /* avoid nasty roots at zero */
    if (smallzero < tolerance && smallzero > -tolerance)
    {
        flt solmin = -linear / quadratic;

        if (solmin > tolerance)
        {
            *hits = solmin;
            rawsphere->hit++;
            return 1;
        } else
            return 0;
    }
    flt discriminant = sqr(linear) - 4 * quadratic * constant;

    if (discriminant < 0)
        return 0;

    quadratic *= 2;
    discriminant = sqrt(discriminant);
    flt    solmax = (-linear + discriminant) / (quadratic);
    flt solmin = (-linear - discriminant) / (quadratic);

    if (solmax < tolerance)
        return 0;               /* maximum is  behind us. Shake it. */

    /* This is an intersection! */
    rawsphere->hit++;
    if (solmin < tolerance)
    {
        *hits = solmax;
        return 1;
    } else
    {
        *hits++ = solmin;
        *hits = solmax;
        return 2;
    }
}

/* Intersect ray with sphere */
int sphere_data::do_intersection(ray * r, flt * hits, int ) const
{
    rawsphere->test--;
    int n =intersect_sphere(center, rsq, r, hits);
    if (n)
	rawsphere->hit--;
    return n;
}

vector  sphere_data::normal(vector x) const
{
    /* calculate the normal.  It is just the direction of the radius */
    return (x - center);
}


void
sphere_data::bound(matrix *trans, bounding_box *bbox)
{
    /* returned minimum and maximum of extent */
    /* bounding volume calculation (ggems) */  
    vector  scal(radius);
    matrix ctm;
    ctm.scale(scal);
    
    matrix tempmatrix;
    tempmatrix.translate(center);
    tempmatrix = tempmatrix * ctm;

    if (trans)
        tempmatrix = (*trans) * (tempmatrix);
    
    ctm = tempmatrix.transpose();

    for (int i = 0; i < 3; i++)
    {
        /* calculate first extremum. */
	flt        u1 = safe_arctangent(ctm[2][i], ctm[0][i]);
	flt     denominator = ctm[0][i] * cos(u1) + ctm[2][i] * sin(u1);
	flt  v1 = safe_arctangent(ctm[1][i], denominator);

        /* second extremum is +/- PI from u1*/
	
	flt u2 = u1  + ((u1 <= 0) ? M_PI :-M_PI);
	/* negative of v1 */
	flt        v2 = -v1;

        /* find and sort extrema locations in this dimension */
        bbox->min[i] =
            ctm[0][i] * cos(u1) * cos(v1) +
            ctm[1][i] * sin(v1) +
            ctm[2][i] * sin(u1) * cos(v1) +
            ctm[3][i];
        bbox->max[i] =
            ctm[0][i] * cos(u2) * cos(v2) +
            ctm[1][i] * sin(v2) +
            ctm[2][i] * sin(u2) * cos(v2) +
            ctm[3][i];

        if (bbox->min[i] > bbox->max[i])
        {
	    flt    tmp = bbox->max[i];
	    bbox->max[i] =bbox->min[i];
	    bbox->min[i] = tmp;
	}
    }
}

void
sphere_data::precompute()
{
    stat->howmuch++;

    rsq = sqr(radius);
    radius = abs(radius);

    prims++;
}
bool
sphere_data::inside(vector x) const
{
    x -= center;
    return (x * x < rsq);
}

void
sphere_data::translate(const vector t)
{
    center += t;
}

void
sphere_data::scale(const vector s)
{
    radius *= abs(s.x);
    center.scale(s);
}

void
sphere_data::rotate(const matrix & m)
{
    center.rotate(m);
}

shape_stats *sphere_data::stat = shape_stats::regist("sphere");
sphere_data::sphere_data(vector c, flt r)
{
    type = SPHERE;
    center = c;
    radius = r;
}

void
sphere_data::print() const
{
#ifdef DEBUG
    dprintf("radius %f ", radius);
    center.print("cntr");
#endif          /* DEBUG */
}

PUBLIC object *
        get_new_sphere_object(vector c, flt r)
{
    object *o = new object;
    o->data = new sphere_data(c,r);
    o->flags = O_CLOSED | O_FINITE_INSIDE  | O_ONLY_UNIFORM_SCALE ;
    o->data->setdaddy(o);

    return o;
}
