#include "extern.hh"
#include "poly.hh"

/****************************************************************
  STURM SEQS
 
  The larger part of this code (the sturm sequence code) is from Art, the
  raytracer from the rendering package Vort. It was written by David G
  Hook, Bernie Kirby and Peter McAree all the credits should go to them.
  You can reach them via <echidna@munnari.oz.au>

  the C++ encapsulation was done by Yours Truly, HWN

 ****************************************************************/

/* max number of iterations */
const int MAXIT=	     	800;
const int MAXP2=		32;

/* smallest relative error we want */
const flt RELERROR=	1.0e-10;

class Sturm_sequence {
    Poly seq[MAX_ORDER];	// seq of polys
    int np;	 		// number of polys

    int evalat0() const;
    int evalatb(flt) const;
    int evalata(flt) const;
    void sbisect(flt min, flt max, int atmin, int atmax, flt * roots) const;
    int csgsbisect(flt min, flt max, int atmin, int atmax, flt * roots) const;
    int numchanges(flt) const;
public:
    Sturm_sequence(Poly const &);
    int solve(flt *, bool);   
};


int
Poly::sturm_solve_rt(flt * roots, bool chkall)
{                   
#if 0				// if the user is stupid, let him.
    /* not really necessary for these polys. */
    if (ord <= 2)
	return solve_rt(roots, chkall);
#endif

     /* avoid degenerate cases. */    
     /* especially avoid intersections close to 0 */
    real_clean();

    while (ord > 0 && coef[ord] == 0.0)
	ord--;

    Sturm_sequence s(*this);
    int r = s.solve(roots, chkall);

    return r;
}


Sturm_sequence::Sturm_sequence(Poly const & tosolve)
{    
    seq[0] = tosolve;


    seq[1] = seq[0];
    seq[1].differentiate();


    /* normalise the leading coefficient. */
    seq[1] /= abs(seq[1].lc());

    Poly  *sp=seq + 2;

    /* construct the rest of the Sturm sequence */
    for (; sp->mod(sp [- 2], sp[ - 1]); sp++)
	{
	    /* reverse the sign and normalise */
	    flt f = -abs(sp->lc());
	    assert (!ISZERO(f));
	    *sp /= f;
	}

    sp->coef[0] = -sp->coef[0]; /* reverse the sign */

    np =  (sp - seq);
}
/**
   build up a sturm sequence for a polynomial in smat, returning the number
   of polynomials in the sequence
  */


int
Sturm_sequence::solve(flt * roots, bool chkall)
{
    /* find sign changes in the sequence at 0 */
    int atmin = evalat0();

     /* find sign changes in the sequence at infinity */
    int    atinf = evalatb(0.0);

    /* the number of sign changes of thepoly in [a,b] is atmin - atinf. This is the important thing of Sturm seqs */
    int nroots = atmin - atinf;
    assert(nroots >= 0);
    /* no roots which are bigger than 0 */
    if (nroots == 0)    
	return 0;


     /* hmm ... I wonder why. */
    atmin = evalata(tolerance);
    nroots = atmin - atinf;

    if (nroots == 0)
	return 0;


    flt min = tolerance;
    flt max = min + 2.0;

     /* find the range where all those signchanges occur, by doubling the interval size. */
    for (int its = 0; its < MAXP2; its++){
	atinf = evalatb(max);

	int         numroots = atmin - atinf;
	/* [min, max] covers all signchanges, therefore it covers all zero's */
	if (numroots == nroots) 
	    break;
	
	max *= 2.0;
    }

    if (nroots == 1 || !chkall){
	
	/* only one root, (or we're only interested in the smallest) */
	sbisect(min, max, atmin, atinf, roots);

	nroots = 1;
    } else {    
	if (csgsbisect(min, max, atmin, atinf, roots) != nroots)
	    warning("can't find roots");
    }

    return nroots;
}

///
int
Sturm_sequence::evalat0() const
{
    int     at0 = 0;

    const Poly *ls = seq;
    for (const Poly *s = seq + 1; s <= seq + np; s++) {
	if (ls->coef[0] * s->coef[0] < 0 || (ls->coef[0] == 0.0 && s->coef[0] != 0.0))
	    at0++;
	ls = s;
    }

    return (at0);
}
/**
  calculate the sign changes in the sturm sequence in smat at 0
 */

 /// calculate the sign changes in the sturm sequence in smat at min value a 
int
Sturm_sequence::evalata(flt a) const
{
    int     ata = 0;
    
    flt    lf = seq[0].eval(a);
    for (const Poly *s = seq + 1; s <= seq + np; s++)
	{
	    flt         f = s->eval(a);
	    if (lf * f < 0 || lf == 0.0)
		ata++;
	    lf = f;
	}

    return (ata);
}


/// calculate the sign changes in the sturm sequence in smat at max value b
int
Sturm_sequence::evalatb(flt b) const
{          
    int         atb = 0;

    if (b == 0.0){	
	for (const Poly *s = seq + 1, *ls = seq; s <= seq + np; s++) {
	    if (ls->lc() * s->lc() < 0
		|| (ls->lc() == 0.0 && s->lc() != 0.0))
		atb++;
	    ls = s;
	}
    } else{
	flt    lf = seq[0].eval(b);
	for (const Poly *s = seq + 1; s <= seq + np; s++){
	    flt f = s->eval(b);
	    if (lf * f < 0 || lf == 0.0)
		atb++;
	    lf = f;
	}
    }
    
    return (atb);
}

///the modified regula-falsi
static bool
modrf(Poly const &thepol, flt a, flt b, flt & val) 
{
#if 0
    const     flt *startcoef = thepol.coef;		
    const     flt *endcoef = &thepol.lc();
#endif
    
    flt fb = thepol.eval(b);
    flt fa = thepol.eval(a);
#if 0
	flt fb = *endcoef;
    flt fa = fb;
    for (const flt *fp = endcoef - 1; fp >= startcoef; fp--){
	fa = a * fa + *fp;
	fb = b * fb + *fp;
    }
#endif
    /* if there is no sign difference the method won't work */
    if (fa * fb > 0.0)
        return false;

    if (abs(fa) < RELERROR)
	{
	    val = a;
	    return true;
	}
    if (abs(fb) < RELERROR)
	{
	    val = b;
	    return true;
	}
    flt    lfx = fa;
    flt       x,fx;

    for (int its = 0; its < MAXIT; its++) {

	x = (fb * a - fa * b) / (fb - fa);
	fx = thepol.eval(x);
	
#if 0
	fx = *endcoef;
	for (const flt *fp = endcoef - 1; fp >= startcoef; fp--)
	    fx = x * fx + *fp;
#endif
	if (abs(x) > RELERROR) {
	    if (abs(fx / x) < RELERROR) {
		val = x;
		return true;
	    }
	} else if (abs(fx) < RELERROR) {
	    val = x;
	    return true;
	}
	if ((fa * fx) < 0) {
	    b = x;
	    fb = fx;
	    if ((lfx * fx) > 0)
		fa /= 2;
	} else {
	    a = x;
	    fa = fx;
	    if ((lfx * fx) > 0)
		fb /= 2;
	}

	lfx = fx;
    }

#ifdef DEBUG
    if (abs(fx) > ALG_TOLERANCE) {

	if (debug_options & DEBUGSELFCHECK)
	    warning("modrf overflow %f %f %f\n", a, b, fx);

    }
#endif          /* DEBUG */

    val = x;

    return true;
}
/** uses the modified regula-falsi method to evaluate the root in interval
  [a,b] of the polynomial described in coef. The root is returned is
  returned in *val. The routine returns zero if it can't converge.
 */

///
void
Sturm_sequence::sbisect(flt min, flt max, int atmin, int atmax, flt * roots) const
{
    for (int its = 0; its < MAXIT; its++){

	flt	mid = (min + max) / 2;     
	int	atmid = numchanges(mid);
	int 	n1 = atmin - atmid;

	if (n1 == 1){
	    /* first try a less expensive technique. */
	    if (modrf(*seq, min, mid, roots[0]))
		return ;

	    /*
    if we get here we have to evaluate the root the
    hard way by using the Sturm sequence. It also means the root is
    repeated an even number of times.
    */

	    // whoahh! shadowing other loop variable
	    for (its = 0; its < MAXIT; its++){

		mid = (min + max) / 2;

		atmid = numchanges( mid);

		if (abs(mid) > RELERROR) {
		    if (abs((max - min) / mid) < RELERROR){
			roots[0] = mid;
			return ;
		    }
		} else if (abs(max - min) < RELERROR){
		    roots[0] = mid;
		    return ;
		}
		if ((atmin - atmid) == 0)
		    min = mid;
		else
		    max = mid;
	    }

#ifdef DEBUG
	    if ( (max - min) > ALG_TOLERANCE &&
		 (debug_options & DEBUGRUNTIME))
		{
		    warning("sbisect: overflow min %f max %f diff %e\n", min, max, max - min);
		}
#endif          /* DEBUG */
	    roots[0] = mid;

	    return ;
	} else if (n1 == 0)
	    min = mid;
	else
	    max = mid;

    }

    roots[0] = min;             /* multiple roots close together */
#ifdef DEBUG
    if (max - min > ALG_TOLERANCE && (debug_options & DEBUGRUNTIME))
        warning("sbisect: overflow min %f max %f diff %e\n", min, max, max - min);
#endif          /* DEBUG */
}
/**
  uses a bisection based on the sturm sequence for the polynomial described
  in seq to isolate the lowest positve root of the polynomial in the
  bounds min and max. The root is returned in *roots.
 */

int
Sturm_sequence::csgsbisect(flt min, flt max, int atmin, int atmax, flt * roots) const
{
    /* assume the worst */
    int nroot = atmin - atmax;
    if (nroot == 1) {

	/* first try a less expensive technique. */
	if (modrf(*seq, min, max, roots[0]))        
	    return 1;
	    
	flt    mid;
	/* if we get here we have to evaluate the root the hard way by
     using the Sturm sequence. It also means the root is repeated an
     even number of times. */
	for (int its = 0; its < MAXIT; its++)
	    {
		mid = (min + max) / 2;

		int atmid = numchanges( mid);

		if (abs(mid) > RELERROR)
		    {
			if (abs((max - min) / mid) < RELERROR)
			    break;
		    } else if (abs(max - min) < RELERROR)
			break;

		if ((atmin - atmid) == 0)
		    min = mid;
		else
		    max = mid;
	    }

	//#ifdef DEBUG
#if 0
	if (its == MAXIT && (debug_options & DEBUGRUNTIME))
	    warning("csgsbisect: overflow min %f max %f diff %e\n", min, max, max - min);

#endif          /* DEBUG */
	roots[1] = roots[0] = mid;
	return 2;
    }

    /* more than one root in the interval, we have to bisect... */
    int    nr = 0;
    flt mid;
    int its = 0;
    for (; its < MAXIT; its++) {
	mid = (min + max) / 2;

	int atmid = numchanges( mid);

	int        n1 = atmin - atmid;
	int     n2 = atmid - atmax;

	if (n1 != 0 && n2 != 0)
	    {
		nr += csgsbisect(min, mid, atmin, atmid, roots);
		nr += csgsbisect(mid, max, atmid, atmax, &roots[n1]);
		break;
	    }
	if (n1 == 0)
	    min = mid;
	else
	    max = mid;
    }

    if (its == MAXIT) {

	//#ifdef DEBUG
#if 0
	if (debug_options & DEBUGRUNTIME)
	    {
		warning("csgsbisect: roots too close together\n");
		warning("csgsbisect: overflow min %f max %f diff %e nroot %d n1 %d n2 %d\n",
			min, max, max - min, nroot, n1, n2);
	    }
#endif          /* DEBUG */

	int n1 = atmin - atmax;
	nr += n1;
	for (int     i = 0; i < n1; i++)
	    roots[i] = mid;
    }
    return nr;
}
/**
  uses a bisection based on the sturm sequence for the polynomial described
  in seq to find the positive roots in the bounds min to max, the roots
  are returned in the roots array in order of magnitude.
 */

#if 0

int
Sturm_sequence::numroots(int &atneg, int &atpos) const
{
    int    atposinf = 0;
    int atneginf = 0;

    /* changes at positve infinity */
    flt    lf = seq[0].lc();

    for (Poly  *s = seq + 1; s <= seq + np; s++)
	{
	    flt     f = s->lc();
	    if (lf == 0.0 || lf * f < 0)
		atposinf++;
	    lf = f;
	}

    /* changes at negative infinity */
    if (seq[0].ord & 1) // whoh?
        lf = -seq[0].lc();
    else
        lf = seq[0].lc();

    for (s = seq + 1; s <= seq + np; s++) {
	if (s->ord & 1)
	    f = -s->lc();
	else
	    f = s->lc();
	if (lf == 0.0 || lf * f < 0)
	    atneginf++;
	lf = f;
    }

    *atneg = atneginf;
    *atpos = atposinf;

    return (atneginf - atposinf);
}
/**
    return the number of distinct real roots of the polynomial described in
    seq.
 */
#endif



int
Sturm_sequence::numchanges(flt a) const
{      
    int     changes = 0;
    flt    lf = seq[0].eval(a);

    for (const Poly  *s = seq + 1; s <= seq + np; s++)
	{
	    flt     f = s->eval(a);
	    if (lf == 0.0 || lf * f < 0)
		changes++;
	    lf = f;
	}

    return changes;
}

/**
 return the number of sign changes in the Sturm sequence in seq at the
  value a.
 */
