/*
   superq.cc -- this should implement superquadrics

   (c) 1993, 1994, 1995 Han-Wen Nienhuys <hanwen@stack.urc.tue.nl>

   this is part of Rayce, a raytracer based on "Ray" by George Kyriazis
   <kyriazis@turing.cs.rpi.edu> Rayce is released under the terms of the
   GNU General Public License, it has no warranty what so ever. For more
   details: see the file "COPYING"
 */


#include "object.hh"
#include "shapes.hh"
#include "proto.hh"
#include "extern.hh"

inline vector SGRAD(vector & g, const vector & X_t, vector n)
{
    g.x = n.x * SGN(X_t.x) * pow(abs(X_t.x), n.x - 1);
    g.y = n.y * SGN(X_t.y) * pow(abs(X_t.y), n.y - 1);
    g.z = n.z * SGN(X_t.z) * pow(abs(X_t.z), n.z - 1);
    return g;
}

inline flt SNORM(vector const & X_t, vector const & n)
{
    return (pow(abs(X_t.x), n.x) +
            pow(abs(X_t.y), n.y) +
            pow(abs(X_t.z), n.z));
}

const int  MAXIT = 200;
#include "ibuf.hh"


bool
superq_data::inside(vector loc) const
{
    return (SNORM(loc, (powvect)) < 1);
}

void
superq_data::print() const
{
#ifdef DEBUG
    powvect.print("exponents");
#endif          /* DEBUG */
}

// should use Newton or Regula Falsi
flt superq_binary_chop(vector const & powvect, ray *r, flt lo_t, flt hi_t) {

    /* use binary chop to get to surface */
    flt mid;
    while (abs(lo_t - hi_t) > ALG_TOLERANCE) {
	mid = (lo_t + hi_t) / 2;
	vector             X_t = r->point(mid);
	
	if (SNORM(X_t, powvect) < 1)
	    hi_t = mid;
	else
	    lo_t = mid;
    }
    
    return mid;
}
/*
 * Intersect a ray with a superquadric.
 */

int superq_data::do_intersection(ray * r, flt * rhits, int flags) const 
{
    bool chkall = flags & CHKALL;
    vector
            lo_grad,
            hi_grad;
    bool    inside;


    /* intersect bounding box */
    flt     bhits[2];
    int         n = intersect_rawbox(vector(-1.01),vector (1.01), r, bhits);
    if (!n)
        return 0;
   
    flt            lo_t,
            hi_t;
    /* determine interval for ray parameter t */
    if (n == 1)
    {
        lo_t = 2 * tolerance;
        hi_t = bhits[0];
    } else
    {
        lo_t = bhits[0];
        hi_t = bhits[1];
    }

    /* are we inside? */
    vector    X_t = r->point(lo_t);
    if (n == 1)
        inside = (SNORM(X_t, powvect) < 1);
    else
        inside = false;

    SGRAD(lo_grad, X_t, powvect);

    if (!inside && lo_grad * r->dir > 0)        /* the shape is behind us. */
        return 0;

    if (inside)
    {
        /* use binary chop to get to surface. We know for sure that hi_t (the box hit) is outside. so an intersection must be
         * between lo_t and hi_t */
	rhits[0] = superq_binary_chop(powvect, r, lo_t,hi_t);
        return 1;

    } else
    {
        /* not inside */
   
	flt middot;
        vector  midgrad;
	
        flt lo_dot = lo_grad * r->dir;

        X_t = r->point(hi_t);
        SGRAD(hi_grad, X_t, powvect);
	flt        hi_dot = hi_grad * r->dir;

        inside = false;

        /* first we try to find the point with (superq_normal, r->dir) == 0 At such a point, the distance to superquadric will be
         * minimal. */
        int             its = MAXIT;
	flt         mid = (lo_t + hi_t) / 2;
        while (abs(lo_t - hi_t) > ALG_TOLERANCE && its--)
        {
            X_t = r->point(mid);

	    SGRAD(midgrad, X_t, powvect);
	    vector v;
            v.x = X_t.x / powvect.x;
            v.y = X_t.y / powvect.y;
            v.z = X_t.z / powvect.z;

            if ((midgrad * v) < 1)
            {
                inside = true;
                break;
            }
            middot = midgrad * r->dir;
            if (middot < 0)
            {
                lo_t = mid;
                lo_grad = midgrad;
                lo_dot = middot;
            } else
            {
                hi_t = mid;
                hi_grad = midgrad;
                hi_dot = middot;
            }
            mid = (lo_t + hi_t) / 2;
        }
        /* minimum distance , and not inside. Exit. */
        if (!inside)
            return 0;

        /* we now know that
         *
         * X_t = pos + mid * dir
         *
         *
         * is inside the superquadric, and pos+lo*dir and pos+hi*dir outside . Now all we have to do is find the intersection
         * inbetween. */



        /* use binary chop to get to surface. */

	rhits[0] = superq_binary_chop(powvect,r, lo_t,mid);
	rhits[1] = superq_binary_chop(powvect,r, mid,hi_t);
        if (!chkall)
            return 1;

        return 2;
    }
}


/*
 * returns the normal to the superq
 */
vector  superq_data::normal(vector loc) const
{
    vector  n;
    SGRAD(n, loc, powvect);

    return n;
}
/* initialize a superq_data */
shape_stats *superq_data::stat = shape_stats::regist("superq");

superq_data::superq_data(vector v)
{
    type = SUPERQ;
    powvect=v;       // a sphere
}



void
superq_data::bound(matrix *trans, bounding_box *bbox)
{
    vector    maxv(1 + BOUNDFUDGE);
    vector    minv(-1 - BOUNDFUDGE);
    do_box_bound(bbox, trans, minv, maxv);
}

void
superq_data::precompute()
{

    stat->howmuch++;
    prims++;
}



/* alloc/init */
PUBLIC object *
        get_new_superq_object(vector v)
{
    object *o = new object;
    o->data = new superq_data(v);
        o->data->setdaddy(o);
    o->flags = O_CLOSED | O_FINITE_INSIDE;

    return o;
}
/* eof */
