/*
   token.cc -- tokenizer to go with YACC/Bison pov scanner

   (c) 1993, 1994, 1995 Han-Wen Nienhuys <hanwen@stack.urc.tue.nl>

   this is part of Rayce, a raytracer based on "Ray" by George Kyriazis
   <kyriazis@turing.cs.rpi.edu> Rayce is released under the terms of the
   GNU General Public License, it has no warranty what so ever. For more
   details: see the file "COPYING"
 */

#include <ctype.h>
#include <string.h>


#include "object.hh"
#include "proto.hh"
#include "extern.hh"
#include "parse.hh"
#include "color.hh"

#if defined MSDOS || defined __MSDOS__ || defined msdos || defined __msdos__
#include "rayparse.hh"
#else
#include "rayparse.tab.hh"
#endif



#define INDICATORCOUNT 25

/*******************************************************************
  FILES
 *******************************************************************/

/* file stack, for INCLUDES */
struct File_stack_entry
{
    FILE   *fp;
    int     lineno;
    File_stack_entry *prev;
    char   *name;
            File_stack_entry(const char *fname);
            ~File_stack_entry();
};

File_stack_entry::File_stack_entry(const char *fname)
{
    name = xstrdup(fname);
    lineno = 1;
    prev = 0;
    fp = open_file(fname, "rt");
    printf("[%s ", fname);
}

File_stack_entry::~File_stack_entry()
{
    delete[]  name;
    printf("]");
    close_file(fp);
}

/**************/
struct File_stack
{
    File_stack_entry *top;

    File_stack();
    void push(const char *s);
    void pop();
    ~File_stack();
};

File_stack::File_stack()
{
    top = 0;
}

File_stack::~File_stack()
{
    assert(!top);
}
void
File_stack::pop()
{
    File_stack_entry *p = top->prev;
    delete  top;
    top = p;
    if (!top)
        printf("<end of scene>\n");
}

void
File_stack::push(const char *fn)
{
    File_stack_entry *p = top;
    top = new File_stack_entry(fn);
    top->prev = p;
}

/****************************************************************/
class Token_input {
    File_stack stack;
    long dotcount;
    void handle_include();
    void get_quotedstring(char *s);
    void get_word(char *cp);
    void eat_junk();
    void pushc(int c);
    int  popc();

    void eat_comment();
    void eat_linecomment();
public:
    int do_lex(YYSTYPE *);
    Token_input(const char *s) {
	stack.push(s);
	dotcount = 0;
    }
    int line_no() { return stack.top->lineno; }
    const char *name() { return stack.top->name; }
    void print_gobs();
};

static Token_input *the_input = 0 ;

/***************************************************************************
 * START/STOP: entry points						   *
 ***************************************************************************/

void
yyerror(const char *s)
{
    printf("\n%s: %d: %s", the_input->name(), the_input->line_no(), s);
    the_input->print_gobs();
    exit(1);
}

void
yywarning(const char *s)
{
    printf("\n%s: %d: warning %s", the_input->name(), the_input->line_no(), s);    
}


/* initialize this module */
void 
Scene::read_from(const char *fname)
{  
    the_input = new Token_input(fname);
    start_parse();
    kill_declarations();
    delete the_input;
}

/*
 * The main tokenizer function. 
 * entry point: do tokenizing, print debug info.
 */
int yylex()
{
    YYSTYPE *lvalp = &yylval;
    
    int code = the_input->do_lex(lvalp);
#ifdef DEBUG
    if (debug_options & DEBUGTOK)

	switch(code) {
    case FLOAT_CONST:
	printf("\nread %f\n", lvalp->fltval);
	break;
    case 0:
	printf("<EOF>");
	break;

    case STRING:
        if (debug_options & DEBUGTOK)
            printf("STRING %s\n", lvalp->stringval);
	break;
    default:
	if (code < 256)
	    printf("CHAR %c \n", code);
	else
	    printf("OTHER %d\n" ,code);
	break;
    }
#endif
    return code;
}

/***************************************************************************
 * PRIVATES								   *
 ***************************************************************************/



/*
 * read the #include "xxx" part
 */
 void
Token_input::handle_include()
{
    char    fn[100];
    get_quotedstring(fn);
    stack.push(fn);
}

/*
 * The main tokenizer function. 
 *
 * get a char: is it an letter ([a-zA-Z_]), then read a word. Is it a
 * keyword? yes: is it include ? yes: handle include, and call yylex()
 * recursively no: return token code no: it's an identifier. Is it new?
 * yes: add a declaration, no: return pointer to info
 */
int
Token_input::do_lex(YYSTYPE *lvalp)
{
    eat_junk();                 /* eat up comments, and whitespace */
    int         c = popc();                 /* look at important char */

    if (isalpha(c) || c == '_')
    {                           /* were looking at a word */
        pushc(c);

	char    s[100];
        get_word(s);
        int t = lookup_keyword(s);

        if (t != -1)
        {                       /* reserved word? */
#ifdef DEBUG
        if (debug_options & DEBUGTOK)
            printf("\nKEYWORD \"%s\"\n", s);
#endif          /* DEBUG */
            if (t == INCLUDE)
            {                   /* includes shouldn't be seen by the parser */
                handle_include();
                return do_lex(lvalp);
            } else
            {
                return t;       /* a real keyword */
            }

        } else
        {
#ifdef DEBUG
	    if (debug_options & DEBUGTOK)
		printf("\nIDENTIFIER \"%s\"\n", s);
#endif          	/* DEBUG */

	    lvalp->dectabptr = lookup_declaration(s);
	    return lvalp->dectabptr->type;
        }
    } else if (isdigit(c) || c == '.')
    {                           /* number */
        pushc(c);
        if (fscanf(stack.top->fp, "%lf", &(lvalp->fltval)) != 1)
            lvalp->fltval = 0;

        return FLOAT_CONST;
    }
    if (c == EOF)
    {
        if (!stack.top)
            return 0;

        stack.pop();
        return do_lex(lvalp);
    }
    if (c == '\"')
    {
        pushc('\"');
        get_quotedstring(lvalp->stringval);
        return STRING;
    }
    /* not a string, not a number, so it's a char; return it. */
    return c;
}

/*
 * read a quoted string like "ad   sfj54%$%$5439yties" gobble the quotes.
 */
void
Token_input::get_quotedstring(char *s)
{

    eat_junk();
    int     c;
    if ((c = popc()) != '\"')
        errormsg("\" expected");

    while ((c = popc()) != EOF && c != '\"')
        *s++ = c;

    if (c != '\"')
        errormsg("\" expected");
    *s = 0;
}

/* handle word; _ is a char too */
void
Token_input::get_word(char *cp)
{

    while (1)
    {
	int             c = popc();
        if (!(isalnum(c) || c == '_'))
        {                       /* OK, end of word */
            *cp = 0;
            pushc(c);
            return ;
        }
        *cp++ = c;
    }
}

const int GOBLEN =20;
void
Token_input::print_gobs()
{
    FILE *fp= stack.top->fp;
    char  s2[GOBLEN+ 1];
    
    int i = fread(s2, sizeof(char), GOBLEN, fp);
    s2[i] = 0;
    long here = ftell(fp);
    fseek (fp, -GOBLEN -i,SEEK_CUR);
    long there = ftell(fp);
    long pre_goblen = here-there - i;

    char s1[GOBLEN+ 1];
    int j = fread(s1, sizeof(char), pre_goblen, fp);
    s1[j] = 0;
    
    printf("\n%s\n<ERROR>\n%s\n", s1, s2);    
}

/* strip comments, spaces etc */
void
Token_input::eat_junk()
{

    while (1)
    {
	int             c = popc();
        if (c == '#' || c == '\r')      /* ignore '#' */
            continue;

        if (c == '/')
        {                       /* is this a comment? */
            c = popc();
            if (c == '*')
                eat_comment();
            else if (c == '/')
                eat_linecomment();
            else
            {                   /* not a comment */
                pushc(c);
                pushc('/');
                return ;
            }
        } else if (!isspace(c))
        {
            pushc(c);
            return ;
        }
    }
}
/* eat a comment */
 void
Token_input::eat_comment()
{

    while (1)
    {
	int             c = popc();
        if (c == EOF)
            return ;

        if (c == '/')
        {
            c = popc();
            if (c != '*')
                continue;
            eat_comment();
        } else if (c == '*')
            if ((c = popc()) == '/')
                return ;
            else
                pushc(c);
    }
}
/* eat a // comment */
 void 
Token_input::eat_linecomment()
{

    while (1) {
	int             c = popc();
        if (c == '\n' || c == EOF)
            return ;
    }
}
/* i'm too lazy to type getc(infile); */
 int
Token_input::popc()
{

    if (!stack.top)
        return EOF;

    int         c = getc(stack.top->fp);

    /* print "happy dots" */
    if (c == '\n')
    {
        if (!(dotcount++ % INDICATORCOUNT))
            printf(".");
        stack.top->lineno++;
    }

    return c;
}
/* push back a char to the current infile */
void 
Token_input::pushc(int c)
{
    if (!stack.top)
        return ;

    if (c == '\n')
        stack.top->lineno--;
    ungetc(c, stack.top->fp);
}
