/*
   torus.cc -- this implements toruses.

 (c) 1993, 1994, 1995 Han-Wen Nienhuys <hanwen@stack.urc.tue.nl>

this is part of Rayce, a raytracer based on "Ray" by George Kyriazis
<kyriazis@turing.cs.rpi.edu> Rayce is released under the terms of the
GNU General Public License, it has no warranty what so ever. For more
details: see the file "COPYING"
 */


#include "object.hh"
#include "shapes.hh"
#include "proto.hh"
#include "extern.hh"
#include "ibuf.hh"
#include "poly.hh"
#include "matrix.hh"

/*
 * a torus eqn looks like factor^2 - stuff = 0.
 */

bool
torus_data::inside(vector loc) const
{
    flt    factor = loc * loc + sqr(Rad) - sqr(rad);
    loc.y = 0;
    flt stuff = 4 * sqr(Rad) * loc * loc;

    return (sqr(factor) - stuff < 0);
}

void
torus_data::print() const
{
#ifdef DEBUG
    dprintf("Major radius %f, Minor radius %f\n", Rad, rad);
#endif          /* DEBUG */
}

/*
 * the bounding was inspired on Graphics Gems II, elliptic torus solver.
 */
int
torus_data::do_intersection(ray *r, flt * rhits, int fl) const 
{
    /* Compute the intersection of the ray with a real minimum bounding sphere. */
    flt     sphere_hits[2];
    int     nhits = intersect_sphere(vector(0), brad, r, sphere_hits);

    if (nhits == 0)
        return 0;

    if (nhits == 2)
    {
        if (sphere_hits[0] > r->maxt)    /* something has been found closer. Early abort. */
            return 0;

        /* Bound the torus by two parallel planes */
	flt             yin = r->pos.y + sphere_hits[0] * r->dir.y;
	flt        yout = r->pos.y + sphere_hits[1] * r->dir.y;

        if ((yin > rad && yout > rad) || (yin < -rad && yout < -rad))
            return 0;

        /* translate the ray along it's direction. This should keep numerical inaccuracies a bit under control. */

        /* does it,  actually? */
	// no.
    }


    /* let the calculations begin! */
    

    /* pos =  baset *  r.dir; pos =  r.pos +  pos; */
    vector pos = r->pos;
    vector dir = r->dir;

    flt    mr = sqr(rad);
    flt  R = sqr(Rad);

        /* the equation is:
         *
         * (z^2 + x^2 + y^2 + R - mr)^2 = 4R(x^2 + z^2)
         *
         * note that R = Rad^2, and mr = rad^2
         *
         * Calculate what happens if the ray is subst in (z^2 + x^2 + y^2 + R - r), which is X * X + R-r. */

    flt cons0 = (pos * pos) + R - mr;
    flt lin0 = 2 * (pos * dir);
    flt quad0 = (dir * dir);

        /* now subst in right part:
         *
         * 4R(x^2+y^2), which is 4R*X * X with X.y = 0 */

        pos.y = dir.y = 0;
    flt        cons1 = R * 4 * (pos * pos);
    flt     lin1 = R * 8 * (pos * dir);
    flt  quad1 = R * 4 * (dir * dir);

        /* now work out
         *
         * (cons0 + t * lin0 + t^2 quad0) ^2 - (cons1 + t * lin1 + t^2 quad1)
	 */
    Poly   tosolve;
    tosolve.coef[0] = sqr(cons0) - cons1;
    tosolve.coef[1] = 2 * cons0 * lin0 - lin1;
    tosolve.coef[2] = sqr(lin0) + 2 * cons0 * quad0 - quad1;
    tosolve.coef[3] = 2 * lin0 * quad0;
    tosolve.coef[4] = sqr(quad0);
    tosolve.ord = 4;    

    /* and solve it. */
    bool chkall = fl & CHKALL;

    if (daddy->flags & O_STURM)
        nhits = tosolve.sturm_solve_rt( rhits, chkall);
    else
        nhits = tosolve.solve_rt( rhits, chkall);

    /* for (i = 0; i < nhits; i++) rhits[i] += ; */


    return nhits;
}
/*
 * The normal has the same direction as a vector from the loc's projection on
 * the torus axis to the intersection point itself. Of course you could
 * also take the gradient of the quartic formula, and then normalize it,
 * but it's a load of work. This is simpler!
 */
vector  torus_data::normal(vector loc) const
{
    vector     heart_proj = loc;
    heart_proj.y = 0;
    heart_proj = (Rad / heart_proj.length()) * heart_proj;

    /* heart_proj now is the projection of loc on the tube axis. */
    vector n	 = loc - heart_proj;

    return n;
}
/* initialize a torus_data */


shape_stats *torus_data::stat = shape_stats::regist("torus");
torus_data::torus_data(flt R, flt r)
{
    type = TORUS;
    Rad = R;
    rad = r;
}



/* bounding volume calculation from Gems III */

/*
 * The bounding volume of a torus is found.  The algorithm performs three
 * passes, one for each dimension.  In each pass the extrema of the
 * surface are found as values of the parameters u and v.  These values
 * are then used to calculate the position of the two points in the
 * current dimension. Finally, these points are sorted to find the minimum
 * and maximum extents. The torus is defined as the rotation of a circle
 * that is perpendicular to the XZ plane.  This circle has radius q, and
 * its center lies in the XZ plane and is rotated about the Y axis in a
 * circle of radius r.  The value of q must be less than that of r.
 */
void
bound_torus(bounding_box *bbox, flt majr, flt minr, const matrix &ictm)
{
    /* cumulative transformation matrix */ 
    flt q = minr;
    flt r = majr;

    if (q > r)
        warning("Weird torus");

    /* Gems do everything transposed. */
    matrix  ctm;
    ctm = ictm.transpose();

    for (int     i = 0; i < 3; i++)
    {

        /* calculate first extremum.  assure that -PI/2 <= v1 <= PI/2 */
	flt        u1 = safe_arctangent(ctm[2][i], ctm[0][i]);
	flt     denominator = ctm[0][i] * cos(u1) + ctm[2][i] * sin(u1);
	flt  v1 = safe_arctangent(ctm[1][i], denominator);

        /* second extremum is +/- PI from u1, negative of v1 */
	
        flt u2 = u1 + ((u1 <= 0) ?  M_PI : -M_PI);
	flt v2 = -v1;

        /* find and sort extrema locations in this dimension */
        bbox->min[i] =
            ctm[0][i] * cos(u1) * (r + q * cos(v1)) +
            ctm[1][i] * sin(v1) * q +
            ctm[2][i] * sin(u1) * (r + q * cos(v1)) +
            ctm[3][i];
        bbox->max[i] =
            ctm[0][i] * cos(u2) * (r + q * cos(v2)) +
            ctm[1][i] * sin(v2) * q +
            ctm[2][i] * sin(u2) * (r + q * cos(v2)) +
            ctm[3][i];
        if (bbox->min[i] > bbox->max[i])
        {
            flt tmp = bbox->max[i];
            bbox->max[i] = bbox->min[i];
            bbox->min[i] = tmp;
        }
    }
}


void
torus_data::bound(matrix *trans, bounding_box *bbox)
{
    matrix  ctm;
    if (trans)
        ctm = *trans;
    else
        ctm.unit();

    bound_torus(bbox, Rad, rad, ctm);
}

void
torus_data::precompute()
{
    brad = Rad + rad + BOUNDFUDGE;
    brad = sqr(brad);
    stat->howmuch++;
    prims++;
}

PUBLIC object *
        get_new_torus_object(flt R, flt r)
{
    object *o = new object;
    o->data = new torus_data(R,r);
    o->flags = O_CLOSED | O_FINITE_INSIDE;
    o->data->setdaddy(o);
    return o;
}
/* eof */









/* not really.... But who actually reads this crap? :) */
