/*
   triangle.c -- handle smooth triangles.

   (c) 1993, 1994, 1995 Han-Wen Nienhuys <hanwen@stack.urc.tue.nl>

   this is part of Rayce, a raytracer based on "Ray" by George Kyriazis
   <kyriazis@turing.cs.rpi.edu> Rayce is released under the terms of the
   GNU General Public License, it has no warranty what so ever. For more
   details: see the file "COPYING"
 */

#include "stats.hh"
#include "object.hh"
#include "triangle.hh"
#include "proto.hh"
#include "extern.hh"
#include "ibuf.hh"


void
triangle_data::scale(const vector s)
{

    for (int     i = 0; i < 3; i++)
	       vertices[i].scale(s);
}

void
triangle_data::rotate(const matrix & rotmat)
{
    for (    int     i = 0; i < 3; i++)
	vertices[i].rotate (rotmat);
}


void
triangle_data::translate(const vector tr)
{
    for (int     i = 0; i < 3; i++)
        vertices[i] += tr;
}

void
triangle_data::print() const
{
#ifdef DEBUG
 
    for (int i = 0; i < 3; i++)
    {
        vertices[i].print("vertex");
    }
    dprintf("projecting along %c\n", project_to+'x');
    linfunc1.print("linfunc1");
    linfunc2.print("linfunc2");
    n.print("normal");
#endif          /* DEBUG */
}


void
smooth_triangle::print() const
{
#ifdef DEBUG
    triangle_data::print();
    
    for (int     i = 0; i < 3; i++)
	normals[i].print("normal");    
#endif          /* DEBUG */
}

bool
triangle_data::all_intersections(ibuf * q, ray * r, int flags, bool &) const
{
    stat->test++;

    /* do plane intersection */
    flt     d = r->dir * n;
    if (ISZERO(d))
        return false;

    dqueue  q_ent;
    q_ent.t = (mov - r->pos * n) / d;

    /* try to get out lazily */
    if (q_ent.t < tolerance || q_ent.t > r->maxt)
        return false;

    
    /* write x-vertex2  as coef1 * (vertex0-vertex1)  + coef2 * (vertex0-vertex2) */

    /* do the intersection. What really happens is the projected version of: */
#if 0
    x = x + r->pos;
    x = q_ent.t * r->dir;
    x = x - vertices[2];

    coef1 = ((x) * (linfunc1));

    if (coef1 < 0)
        return false;

    coef2 = ((x) * (linfunc2));
    if (coef2 < 0)
        return false;
#endif


    int u = (project_to + 1)%3;
    int v = (project_to + 2)%3;
    int w = project_to ;

    vector  x;
    x[w] = 0.0;
    x[u] = r->pos[u] + q_ent.t * r->dir[u];
    x[v] = r->pos[v] + q_ent.t * r->dir[v];

    /* bounding box test. Inspired on GGems */
    if ((x[u] < bmin[u]) || (x[u] > bmax[u]) || (x[v] < bmin[v]) || (x[v] > bmax[v]))
	return false;

    x[u] = x[u] - vertices[2][u];
    x[v] = x[v] - vertices[2][v];
    flt coef1 = x[u] * linfunc1[u] + x[v] * linfunc1[v];
    if (coef1 < 0)
	return false;
    
    flt coef2 = x[u] * linfunc2[u] + x[v] * linfunc2[v];

    if (coef2 < 0)
	return false;

    if (coef1 + coef2 > 1)
        return false;
    else
    {
        q_ent.obj = daddy;
        q->add_to(flags, q_ent);
        stat->hit++;
        return true;
    }
}

vector  triangle_data::normal(vector ) const
{ 
     return n;
}



shape_stats *triangle_data::stat = shape_stats::regist("triangle");

triangle_data::triangle_data(vector v1,vector v2,vector v3)
{
    type = TRIANGLE;
    vertices[0]=v1;
    vertices[1]=v2;
    vertices[2]=v3;
    linfunc1.set(0);
    linfunc2.set(0);
    project_to = 0;
}


smooth_triangle::smooth_triangle(vector v1,vector n1,vector v2,vector n2,vector v3,vector n3)
    :triangle_data(v1,v2,v3)
{
    normals = new vector[3];
    normals[0]=n1;
    normals[1]=n2;
    normals[2]=n3;
}

void
triangle_data::bound(matrix *, bounding_box *bbox)
{
    bbox->set_empty();

    for (int i = 0; i < 3; i++)
        bbox->add(vertices[i]);
    
    bmin = bbox->min;
    bmax = bbox->max;
}

void
triangle_data::precompute()
{
    bmin.set(-INFTY);
    bmax.set( INFTY);
    stat->howmuch++;
    vector    bas1 = vertices[0] - vertices[2];
    vector    bas2 = vertices[1] - vertices[2];

    n = bas1 ^ bas2;
    flt    normlen = n.length();
    if (normlen < EPSILON)
        warning("degenerate triangle");
    else
        n /= normlen;

    mov = n* vertices[2];

    /* project along which axis ? */
    project_to = n.dominant();


    /* we want to express x in a_1 x_1 + a_2 x_2 + a_3 x_3, with a_1 +
    a_2+ a_3 = 1 this is equivalent to (x - x_3) = a_1 ( x_1 - x_3) +
    a_2 (x_2 -x_3) So we have to express x-x_3 in the basis {
    (x_1-x_3), (x_2-x_3) }

    The theory of finite dimensional vectorspaces states, that the
    coefficients a_1 and a_2 can be found by taking an improduct of
    suitable linfunc1 and linfunc2 with x - x_3. These linfunc[12] are
    computed here:

    since this is a 2D problem, we might as well project the intersection
    and the vertices onto one of the coordinate planes.

    first we compute orth, a vector orthogonal to bas1 = x1 - x3
    */

    // numerically stable?
    bas1[ project_to] = bas2[ project_to] =  0;
    flt alpha = -(bas1 * bas2) / (bas1 * bas1);
    vector orth = alpha * bas1;
    orth += bas2;

    linfunc2 = orth / (orth * orth);
    linfunc1 = alpha * linfunc2;
    linfunc1 += bas1 / (bas1 * bas1);

    prims++;
}

/* alloc/init */
PUBLIC object *
        get_new_triangle_object(vector v1,vector v2,vector v3)
{
    object *o = new object;
    o->data = new triangle_data(v1,v2,v3);
    o->data->setdaddy(o);
    return o;
}


/* eof */









/* alloc/init */
PUBLIC object *
        get_new_smooth_triangle_object(vector v1,vector n1,vector v2,vector n2,vector v3,vector n3)
{
    object *o = new object;
    smooth_triangle *p = new smooth_triangle(v1,n1,v2,n2,v3,n3);
    o->data = p;
    o->data->setdaddy(o);
    return o;
}

smooth_triangle::~smooth_triangle()
{
    delete[] normals;
}

void
smooth_triangle::scale(vector s)
{
    triangle_data::scale(s);
    vector  invs(1 / s.x, 1 / s.y, 1 / s.z); 
    for (int i =0; i < 3 ; i++){
            /* preserve length */
	    flt                 len = (normals[i]).length();
            normals[i].scale(invs);
            normals[i] /= (normals[i]).length();
            normals[i] *= len;
	}
}

void
smooth_triangle::rotate(const matrix & rotmat)
{
    triangle_data::rotate(rotmat);
    for (int     i = 0; i < 3; i++) {
            normals[i].rotate (rotmat);
    }
}


smooth_triangle::smooth_triangle(const smooth_triangle & src)
     : triangle_data(src)
{
    normals = new vector[3];
    normals[0] = src.normals[0];
    normals[1] = src.normals[1];
    normals[2] = src.normals[2]; 
}





vector  smooth_triangle::normal(vector loc) const
{
    loc = loc - vertices[2];
    loc [project_to]  = 0;

    /* decompose into base formed by two edges */
    flt     coef1 = loc * linfunc1;
    flt     coef2 = loc * linfunc2;
    flt     coef3 = 1 - coef1 - coef2;

    vector  outn = coef1 * normals[0];
    outn += coef2 * normals[1];
    outn += coef3 * normals[2];
    return n;
}

/* not really.... But who actually reads this crap? :) */
